@@ -464,8 +464,8 @@ class QgridWidget(widgets.DOMWidget):
     _model_name = Unicode('QgridModel').tag(sync=True)
     _view_module = Unicode('qgrid').tag(sync=True)
     _model_module = Unicode('qgrid').tag(sync=True)
-    _view_module_version = Unicode('1.0.3').tag(sync=True)
-    _model_module_version = Unicode('1.0.3').tag(sync=True)
+    _view_module_version = Unicode('1.0.4').tag(sync=True)
+    _model_module_version = Unicode('1.0.4').tag(sync=True)
 
     _df = Instance(pd.DataFrame)
     _df_json = Unicode('', sync=True)
@@ -985,7 +985,7 @@ class QgridWidget(widgets.DOMWidget):
             return
         else:
             if col_info['type'] == 'any':
-                unique_list = col_series.dtype.categories
+                unique_list = col_series.cat.categories
             else:
                 if col_name in self._sorted_column_cache:
                     unique_list = self._sorted_column_cache[col_name]
